import os, warnings, copy, random
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from statannot import add_stat_annotation

import seaborn as sns
import sklearn.linear_model as lm
import sklearn.model_selection as mod_sel
import sklearn.metrics as met
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import auc, roc_curve
from scipy import stats
from sklearn.svm import SVC
from sklearn.ensemble import GradientBoostingClassifier


def evaluate_roc_auc( X, y, vlst, classifier, param_grid, 
                      NCV = 5, target_cls = 'Cancer', title_t = None,
                      fpr_th = 0.01, tpr_th = 0.99, 
                      bsr = 0.8, bsn = 40, vbs = 0, show_std = False):

    cv = mod_sel.StratifiedKFold(n_splits=NCV)
    plt.figure(figsize=(4, 4), dpi=100)
    legend_str = []

    df_log_x = X
    
    sens = []
    sens_all = []
    spec = []
    spec_all = []
    auc_all = []
    groups = []
    auc_mean = []
    auc_std = []
    aucs_lst = []
    
    colors_mean = ['blue', 'red', 'green', 'gold', 'navy', 'orange', 'purple', 'firebrick', 'royalblue']
    mean_fpr_lst = []
    tprs_lower_lst = []
    tprs_upper_lst = []
    for k, key in enumerate(list(vlst.keys())):
        
        df_a = df_log_x[vlst[key]]
    
        ## (4) Search grid to find best parameters
        gs = mod_sel.GridSearchCV(classifier, param_grid, cv=cv, 
                                  scoring='balanced_accuracy', 
                                  n_jobs=NCV, verbose = vbs)
    
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")    
            gs.fit(df_a,y)
    
        ## (5) Check best score and Get best paramter
        best_param = gs.best_params_
        print("Best Scroe: " + str(round(gs.best_score_,4)) + " with parameters ", end='' )
        print(gs.best_params_)
    
        ## Get the best classifiers
        classifier = gs.best_estimator_
    
        idx_all = list(df_a.index.values)
        tprs = []
        aucs = []
        coef_lst = []
        mean_fpr = np.linspace(0, 1, 100)
    
        sens_tmp = []
        spec_tmp = []
        for i in range(bsn):
            idx_sel = random.sample(idx_all, int(len(idx_all)*bsr))
            cv = mod_sel.StratifiedKFold(n_splits=NCV)
            
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")    
                y_dec_met = mod_sel.cross_val_predict( classifier, df_a.loc[idx_sel,:], 
                                                       y[idx_sel], cv=cv, 
                                                       method = 'decision_function', 
                                                       verbose = vbs, n_jobs = NCV )
        
            # Compute ROC curve and ROC area for each class
            # cls_lst = list(set(y))
            if classifier.classes_[0] == target_cls:
                y_dec_met = -y_dec_met
        
            fpr, tpr, _ = roc_curve(y[idx_sel].ravel(), y_dec_met.ravel(), pos_label = target_cls)
            roc_auc = auc(fpr, tpr)
        
            sn = 0
            for j, fr in enumerate(fpr):
                if fr <= fpr_th:
                  sn = tpr[j]
            sp = 0
            for j, tp in enumerate(tpr):
                if tp <= tpr_th:
                  sp = 1-fpr[j]

            sens_tmp.append(sn)
            sens_all.append(sn)
            spec_tmp.append(sp)
            spec_all.append(sp)
            auc_all.append(roc_auc)
            groups.append(key)
            
            title = "%s (AUC[%i] = %0.3f)" % (key, i, roc_auc)
    
            if False: # i < 6:
                legend_str.append(title)
                plt.plot(fpr, tpr, label=title, lw = 0.3 ) 
            
            interp_tpr = np.interp(mean_fpr, fpr, tpr)
            tprs.append(interp_tpr)
            aucs.append(roc_auc)
    
        aucs_lst.append(aucs)
        sens.append(np.mean(sens_tmp))
        spec.append(np.mean(spec_tmp))
        mean_tpr = np.mean(tprs, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(aucs)
        plt.plot(mean_fpr, mean_tpr, color=colors_mean[k],
                label='%s (AUC: %0.3f $\pm$ %0.3f)' % (key, mean_auc, std_auc),
                lw=2, alpha=.8)
        auc_mean.append(mean_auc)
        auc_std.append(std_auc)
        
        std_tpr = np.std(tprs, axis=0)
        tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
        tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
        
        if show_std:
            mean_fpr_lst.append(mean_fpr)
            tprs_lower_lst.append(tprs_lower)
            tprs_upper_lst.append(tprs_upper)
            # plt.fill_between(mean_fpr, tprs_lower, tprs_upper, color=colors_mean[k], alpha=.2,
            #                  label=r'%s $\pm$ 1 std. dev.' % key)
    
    plt.plot([0, 1], [0, 1], color="gray", lw=2, linestyle="--", label='Chance', alpha=.8)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel("False Positive Rate = FP/(FP+TN)") #, fontsize = 11)
    plt.ylabel("True Positive Rate = TP/(TP+FN)") #, fontsize = 11)    
    #plt.title("ROC for %s classification" % target_cls )
    if title is not None: plt.title(title_t )
    # plt.legend(loc="lower left", bbox_to_anchor = (1.05, -0.075), fontsize = 9)
    plt.legend(loc="lower right", bbox_to_anchor = (1, 0), fontsize = 9)

    if show_std:
        k = 0
        for mean_fpr, tprs_lower, tprs_upper in zip(mean_fpr_lst, tprs_lower_lst, tprs_upper_lst):
            plt.fill_between(mean_fpr, tprs_lower, tprs_upper, color=colors_mean[k], alpha=.2,
                             label=r'%s $\pm$ 1 std. dev.' % key)
            k += 1
    
    plt.show()
    df_res = pd.DataFrame({'Case': groups, 'AUC': auc_all, 
                           'Sensitivity': sens_all, 'Specificity': spec_all})

    return df_res


def plot_pct_box( df_pct, sg_map, ncols, figsize = None, dpi = 100,
                  title = None, title_y_pos = 1.05, title_fs = 14, 
                  label_fs = 11, tick_fs = 10, xtick_rot = 0, xtick_ha = 'center', group_order = None,
                  annot = True, annot_ref = None, annot_fmt = 'simple', annot_fs = 10, 
                  ws_hs = (0.3, 0.3), ylabel = None, stripplot = True ):

    title2_fs = None
    if title2_fs is None:
        title2_fs = title_fs - 2
    df = df_pct.copy(deep = True)
    # nr, nc = nr_nc
    nc = ncols
    nr = int(np.ceil(df.shape[1]/nc))
    if figsize is None:
        figsize = (3*nc,3*nr)
    else:
        figsize = (figsize[0]*nc,figsize[1]*nr)
        
    ws, hs = ws_hs
    fig, axes = plt.subplots(figsize = figsize, dpi=dpi, nrows=nr, ncols=nc)
    plt.subplots_adjust(wspace=ws, hspace=hs)

    if title is not None:
        fig.suptitle('%s' % title, x = 0.5, y = title_y_pos, fontsize = title_fs, ha = 'center')

    items = list(df.columns.values)

    df['Group'] = list(df.index.values)
    df['Group'].replace(sg_map, inplace = True)

    lst = df['Group'].unique()

    for kk, item in enumerate(items):
        plt.subplot(nr,nc,kk+1)
        b = df[item].isnull()
        pcnt = df.loc[~b,'Group'].value_counts()
        b1 = pcnt > 1
        group_sel = pcnt.index.values[b1].tolist()

        if len(group_sel) > 1:
            b = (~b) & df['Group'].isin(group_sel)
            
            lst = df.loc[b, 'Group'].unique()
            lst_pair = []
            if annot_ref in lst:
                for k, l1 in enumerate(lst):
                    if l1 != annot_ref:
                        lst_pair.append((annot_ref, l1))
            else:
                for k, l1 in enumerate(lst):
                    for j, l2 in enumerate(lst):
                        if j <  k:
                            lst_pair.append((l1, l2))
        
            ax = sns.boxplot(data = df.loc[b,:], x = 'Group', y = item, order = group_order)
            if stripplot:
                sns.stripplot(x='Group', y=item, data=df.loc[b,:], jitter=True, 
                              color="black", order = group_order)
            if kk%nc == 0: 
                ax.set_ylabel(ylabel, fontsize = label_fs)                
            else: 
                ax.set_ylabel(None)
                
            if kk >= nc*(nr-1): 
                ax.set_xlabel('Condition', fontsize = label_fs)
            else: 
                ax.set_xlabel(None)
            ax.set_title(item, fontsize = title2_fs)
            
            if kk < (nr*nc - nc):
                # plt.xticks([])
                plt.xticks(rotation = xtick_rot, ha = xtick_ha, fontsize = tick_fs)
                plt.yticks(fontsize = tick_fs)
            else:
                plt.xticks(rotation = xtick_rot, ha = xtick_ha, fontsize = tick_fs)
                plt.yticks(fontsize = tick_fs)
    
            if annot:
                add_stat_annotation(ax, data=df.loc[b,:], x = 'Group', y = item, 
                                    order = group_order,
                            box_pairs=lst_pair, loc='inside', fontsize = annot_fs,
                            test='t-test_ind', text_format=annot_fmt, verbose=0)
            #'''
    
    if (len(items) < (nr*nc)) & (nr > 1):
        for k in range(nr*nc - len(items)):
            axes[nr-1][len(items)%nc + k].axis("off")
        
    plt.show()
    return 
